"""
Wrapper functions for json files.

Behaviour of the JSON writer is changed to:
    - Automatically converts JSON incompatible data types to lists
    (pathlib.Path, numpy arrays, torch tensors, jax tensors) instead of raising errors
    - Does not indent each list element as a new line

Possible improvements:
    - Refactor the "file_or_io" handling into a context manager

"""
import io
import json
import lzma
import os
from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Iterable, Sequence, List

from packg.iotools.compressed import load_xz, get_compressor, get_decompressor
from packg.iotools.file_reader import open_file_or_io, read_text_from_file_or_io
from packg.iotools.jsonext_encoder import CustomJSONEncoder
from packg.typext import PathOrIO, PathType, PathTypeCls


def load_json(file_or_io: PathOrIO, verbose: bool = False, encoding: str = "utf-8") -> Any:
    """Load data from json file or file object"""
    start_timer = timer()
    if verbose:
        try:
            file_len = f"{Path(file_or_io).stat().st_size / 1024 ** 2:.1f} MB"
        except Exception:
            try:
                file_len = f"{len(file_or_io) / 1024 ** 2:.1f} M chars"
            except Exception:
                file_len = f"unknown"
        print(f"Load json file {file_or_io} with size {file_len}.")
    data_str = read_text_from_file_or_io(file_or_io, encoding=encoding)
    try:
        obj = loads_json(data_str)
    except Exception as e:
        raise RuntimeError(f"Error loading json file {file_or_io}") from e
    if verbose:
        print(f"Loaded json file {file_or_io} in {timer() - start_timer:.3f} seconds")
    return obj


def loads_json(s: str) -> Any:
    """Load data from json string"""
    return json.loads(s)


def load_jsonl(file_or_io: PathOrIO, encoding: str = "utf-8") -> List[Any]:
    """Load data from jsonl (list of json strings) file or file object.

    Notes:
        This is inefficient for large files. In that case, iterate the file lines
        and use loads_json on each line.
    """
    data_str = read_text_from_file_or_io(file_or_io, encoding=encoding)
    try:
        data = loads_jsonl(data_str)
    except Exception as e:
        raise RuntimeError(f"Error loading json file {file_or_io} (see reraised error)") from e
    return data


def loads_jsonl(s: str) -> List[Any]:
    data = []
    for i, line in enumerate(s.splitlines()):
        try:
            data.append(loads_json(line))
        except Exception as e:
            raise RuntimeError(f"Error loading json line {i}: {line}") from e
    return data


def dump_json(
    obj: Any,
    file_or_io: PathOrIO,
    ensure_ascii: bool = False,
    check_circular: bool = False,
    allow_nan=False,
    indent=None,
    separators=None,
    sort_keys=False,
    verbose: bool = True,
    create_parent=False,
    float_precision=None,
    custom_format=True,
    encoding="utf-8",
    overwrite=True,
) -> None:
    """Write data to json file or file object using the custom json encoder"""
    start_timer = timer()
    if not overwrite and isinstance(file_or_io, PathTypeCls) and Path(file_or_io).is_file():
        if verbose:
            print(f"File already exists and overwrite is False: {Path(file_or_io).as_posix()}")
        return

    with open_file_or_io(
        file_or_io, mode="wt", encoding=encoding, create_parent=create_parent
    ) as fh:
        if indent is None and separators is None:
            separators = (",", ":")

        kwargs = dict(
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            indent=indent,
            separators=separators,
            sort_keys=sort_keys,
        )
        if custom_format:
            kwargs.update(
                dict(
                    cls=CustomJSONEncoder,
                    float_precision=float_precision,
                )
            )
        json.dump(obj, fh, **kwargs)

    if verbose:
        print(f"Wrote json file {file_or_io} in {timer() - start_timer:.3f} seconds")


def dumps_json(
    obj: Any,
    ensure_ascii: bool = False,
    check_circular: bool = False,
    allow_nan=False,
    indent=None,
    separators=None,
    sort_keys=False,
    float_precision=None,
) -> str:
    """Write data to json string using the custom json encoder"""
    return json.dumps(
        obj,
        cls=CustomJSONEncoder,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        indent=indent,
        separators=separators,
        sort_keys=sort_keys,
        float_precision=float_precision,
    )


def dump_jsonl(
    data: Iterable[Any],
    file_or_io: PathOrIO,
    verbose: bool = True,
    create_parent=False,
    encoding="utf-8",
) -> None:
    """Write lines of data to jsonl (list of json strings) file or file object
    using the custom json encoder"""
    start_timer = timer()
    err_msg = f"data must be a list/sequence but is {type(data)}"
    assert not isinstance(data, str), err_msg
    assert isinstance(data, Sequence), err_msg

    with open_file_or_io(file_or_io, "wt", encoding=encoding, create_parent=create_parent) as fh:
        for d in data:
            fh.write(f"{dumps_json(d)}\n")

    if verbose:
        print(f"Wrote jsonl file {file_or_io} in {timer() - start_timer:.3f} seconds")


def dumps_jsonl(data: Iterable[Any]) -> str:
    sio = io.StringIO()
    dump_jsonl(data, sio, verbose=False)
    return sio.getvalue()


def load_json_xz(file: PathType, verbose: bool = False, encoding: str = "utf-8") -> Any:
    start_timer = timer()
    file = Path(file)
    data_str = load_xz(file, encoding=encoding)
    try:
        obj = loads_json(data_str)
    except Exception as e:
        raise RuntimeError(f"Error loading json file {file}") from e
    if verbose:
        print(f"Loaded json file {file} in {timer() - start_timer:.3f} seconds")
    return obj


def load_json_compressed(
    file: PathType, compressor_name: str, verbose: bool = False, encoding: str = "utf-8"
) -> Any:
    start_timer = timer()
    file = Path(file)
    decompressor = get_decompressor(compressor_name)
    data_bytes = Path(file).read_bytes()
    data_str = decompressor.decompress(data_bytes).decode(encoding)
    try:
        obj = loads_json(data_str)
    except Exception as e:
        raise RuntimeError(f"Error loading json file {file}") from e
    if verbose:
        print(f"Loaded json file {file} in {timer() - start_timer:.3f} seconds")
    return objo


def dump_json_xz(
    obj: Any,
    file: PathType,
    ensure_ascii: bool = False,
    check_circular: bool = False,
    allow_nan=False,
    indent=None,
    separators=None,
    sort_keys=False,
    verbose: bool = True,
    create_parent=False,
    float_precision=None,
    encoding="utf-8",
) -> None:
    start_timer = timer()
    file = Path(file)
    if create_parent:
        os.makedirs(file.parent, exist_ok=True)
    with lzma.open(file, "wt", encoding=encoding) as fh:
        try:
            dump_json(
                obj,
                fh,
                ensure_ascii=ensure_ascii,
                check_circular=check_circular,
                allow_nan=allow_nan,
                indent=indent,
                separators=separators,
                sort_keys=sort_keys,
                verbose=verbose,
                create_parent=create_parent,
                float_precision=float_precision,
            )
        except Exception as e:
            raise RuntimeError(f"Error dumping json xz file {file}") from e

    if verbose:
        print(f"Wrote json xz file {file} in {timer() - start_timer:.3f} seconds")



"""
Remove build artifacts and other clutter from a directory.
"""
import shutil
from pathlib import Path
from typing import Optional

from loguru import logger

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from typedparser import VerboseQuietArgs, add_argument, TypedParser
from attrs import define


@define
class Args(VerboseQuietArgs):
    base_dir: Optional[Path] = add_argument(
        shortcut="-b", type=str, help="Dir to clean", default="."
    )
    write: bool = add_argument(shortcut="-w", action="store_true", help="Confirm removing files")
    clean_packages: bool = add_argument(
        shortcut="-p", action="store_true", help="remove package builds and dist"
    )
    clean_pycache: bool = add_argument(
        shortcut="-c", action="store_true", help="remove all pycache and pytests cache"
    )
    clean_all: bool = add_argument(shortcut="-a", action="store_true", help="remove everything")


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    base_dir = args.base_dir.resolve()
    logger.info(f"Running in {base_dir}. Use -v to show all deletions")

    def rmtree_verbose(path):
        logger.debug(f"removing path {path}")
        if args.write:
            shutil.rmtree(path, ignore_errors=True)

    glob_strs = []
    if args.clean_packages or args.clean_all:
        glob_strs = ["**/build", "**/dist", "**/*.egg-info"]

    if args.clean_pycache or args.clean_all:
        glob_strs += ["**/__pycache__", "**/.pytest_cache"]

    if args.clean_all:
        glob_strs += ["**/.ipynb_checkpoints"]

    logger.info(f"Searching for {glob_strs}")
    n_removed = 0
    for glob_str in glob_strs:
        for file in base_dir.glob(glob_str):
            rmtree_verbose(file)
            n_removed += 1

    logger.info(f"Removed {n_removed} files or dirs")
    if len(glob_strs) == 0:
        logger.warning("No operation specified")

    if not args.write:
        logger.warning("Test run (no -w), did not do anything.")


if __name__ == "__main__":
    main()



"""
Create a bash autocompletion script for a package or several packages .

Note: If it does not find a module, maybe __init__.py is missing.

see  /usr/share/bash-completion/bash_completion complete -F _longopt complete
to figure out how to complete the 2nd argument same as cat would complete the 1st argument.

"""

import importlib_resources
from attrs import define
from loguru import logger
from pathlib import Path
from typing import Optional

from packg.log import SHORTEST_FORMAT, configure_logger, get_logger_level_from_args
from packg.packaging import (
    create_bash_autocomplete_script,
    FILEDIR_AUTOCOMPLETE,
    create_new_bash_autocomplete_script,
)
from typedparser import VerboseQuietArgs, add_argument, TypedParser


@define
class Args(VerboseQuietArgs):
    packages: str = add_argument(shortcut="-p", type=str, default="packg,visiontext")
    target_script: Optional[Path] = add_argument(
        shortcut="-t",
        type=Path,
        help="Target script. If not given, will write to ./autocomplete.sh",
    )
    run_dir: Optional[str] = add_argument(
        shortcut="-r", type=str, help="Only create autocompletion for this directory"
    )
    command_name: str = add_argument(
        shortcut="-c", type=str, default="py", help="Command name to call python"
    )


def main():
    parser = TypedParser.create_parser(Args, description=__doc__)
    args: Args = parser.parse_args()
    configure_logger(level=get_logger_level_from_args(args), format=SHORTEST_FORMAT)
    logger.info(f"{args}")

    packages = args.packages.split(",")
    autocomplete_script = create_new_bash_autocomplete_script(packages, run_dir=args.run_dir, command_name=args.command_name)
    if args.target_script is None:
        args.target_script = "autocomplete.sh"
    Path(args.target_script).write_text(
        "\n".join([FILEDIR_AUTOCOMPLETE, autocomplete_script]), encoding="utf-8"
    )
    logger.info(f"Created: {args.target_script}")
    logger.debug(autocomplete_script)


if __name__ == "__main__":
    main()



"""
Shortcut to create a pathspec from a list of patterns with either .gitignore or regex syntax.

https://pypi.org/project/pathspec/#description

"""
from pathlib import Path
from typing import Iterable, Union, List, Optional

import pathspec
from attr import define

from packg.typext import PathType
from typedparser import add_argument


def make_git_pathspec(patterns: List[str]) -> pathspec.PathSpec:
    """
    >>> spec1: pathspec.PathSpec = make_git_pathspec(["hello*world"])
    >>> print(spec1.match_file("hello_world"))
    True
    >>> print(spec1.match_file("hello_world.txt"))
    False

    Args:
        patterns:

    Returns:

    """
    spec = pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, patterns)
    return spec


def make_regex_pathspec(patterns: List[str]) -> pathspec.PathSpec:
    spec = pathspec.PathSpec.from_lines(pathspec.RegexPattern, patterns)
    return spec


def make_pathspec(patterns: List[str], regex_mode: bool = False) -> pathspec.PathSpec:
    """
    Create PathSpec object that can be used to match or search files.

    Args:
        patterns: separated list of patterns to search for
        regex_mode: False (default) = .gitignore syntax, True = regex

    Returns:
        PathSpec object
    """
    if regex_mode:
        return make_regex_pathspec(patterns)
    else:
        return make_git_pathspec(patterns)


def apply_specs_to_paths(
    paths: List[PathType],
    include_git: Optional[List[str]] = None,
    include_regex: Optional[List[str]] = None,
    exclude_git: Optional[List[str]] = None,
    exclude_regex: Optional[List[str]] = None,
    gitginore_file: Optional[PathType] = None,
) -> Iterable[Path]:
    if include_git is not None and len(include_git) > 0:
        spec = make_git_pathspec(include_git)
        paths = spec.match_files(paths)
    if include_regex is not None and len(include_regex) > 0:
        spec = make_regex_pathspec(include_regex)
        paths = spec.match_files(paths)
    if exclude_git is not None and len(exclude_git) > 0:
        spec = make_git_pathspec(exclude_git)
        paths = spec.match_files(paths, negate=True)
    if exclude_regex is not None and len(exclude_regex) > 0:
        spec = make_regex_pathspec(exclude_regex)
        paths = spec.match_files(paths, negate=True)
    if gitginore_file is not None:
        spec = make_git_pathspec(gitginore_file.read_text(encoding="utf-8").splitlines())
        paths = spec.match_files(paths, negate=True)
    return paths


@define(slots=False)
class PathSpecArgs:
    exclude_git: List[str] = add_argument(
        shortcut="-x", default=[], action="append", help="Git-like pathspec list to exclude files."
    )
    exclude_regex: List[str] = add_argument(
        shortcut="-X", default=[], action="append", help="Regex pathspec list to exclude files."
    )
    include_git: List[str] = add_argument(
        shortcut="-i", default=[], action="append", help="Git-like pathspec list to include files."
    )
    include_regex: List[str] = add_argument(
        shortcut="-I", default=[], action="append", help="Regex pathspec list to include files."
    )
    gitignore_file: Optional[Path] = add_argument(
        shortcut="-g", type=str, default=None, help="Gitignore file for matching files"
    )


class PathSpecWithConversion:
    """
    Wrapper for pathspec.PathSpec with argument conversion to and from pathlib.Path

    Usage:
        matches = spec.match_tree("path/to/dir")
        matches = spec.match_files(file_paths)
        is_matched = spec.match_file("path/to/file")

    Args:
        patterns: separated list of patterns to search for
        regex_mode: False (default) = .gitignore syntax, True = regex
        output_mode: How to return the matched files. One of "str", "path", "str_posix"
    """

    CONVERTERS = {
        "str": str,
        "path": Path,
        "str_posix": lambda input_path: Path(input_path).as_posix(),
    }

    def __init__(
        self,
        patterns: List[str],
        regex_mode: bool = False,
        input_mode: str = "str",
        output_mode: str = "path",
    ):
        self.patterns = patterns
        self.spec = make_pathspec(patterns, regex_mode=regex_mode)
        self.converter_in = self.CONVERTERS[input_mode]
        self.converter_out = self.CONVERTERS[output_mode]

    def match_tree(self, path: Union[str, Path]) -> Iterable[Path]:
        """Matches all files in path recursively, automatically excludes folders."""
        path = self.converter_in(path)
        for file in self.spec.match_tree(path):
            yield self.converter_out(file)

    def match_files(self, paths: List[Union[str, Path]]) -> Iterable[Path]:
        """Yields all entries in paths that match, does not exclude folders."""
        paths = [self.converter_in(path) for path in paths]
        for file in self.spec.match_files(paths):
            yield self.converter_out(file)

    def match_file(self, path: Union[str, Path]) -> bool:
        """Matches this single path, does not exclude folders."""
        return self.spec.match_file(self.converter_in(path))
